(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uindex_set_types.mli,v 1.1 2013/06/04 17:11:41 fclement Exp $ *)

(** {3 The private type of set of multi-indices}

 A mono-dimensional range is a finite interval of natural numbers; it is a
 subset of the set of naturals {b N}.
 A multi-dimensional range of size [n] is the cartesian product of [n]
 mono-dimensional ranges; it is a subset of {b N}^[n].

 An index is a natural number meant to be used as the component number of an
 array, or a vector. A mono-dimensional range is a particular set of indices.

 A multi-dimensional index, or multi-index, of size [n] is an [n]-tuple of
 indices, i.e. an element of {b N}^[n].
 A set of multi-indices may be either a multi-dimensional range, or an explicit
 set of multi-indices, or the union of two set of multi-indices. *)

type 'a multi_d = 'a list;;
(** The multi-dimensionality is represented by lists. *)

(** {6 Range of natural numbers} *)

type natural_range = private
  {
    left : Unatural.t;
    right : Unatural.t;
  }
;;
(** The type of natural ranges.
 [{ left = n; right = p; }] represents the range of natural numbers from
 [n] (inclusive) to [p] (exclusive), i.e. the set [{n, n+1, ..., p-1}].
 Obviously, we have [n <= p].
 Moreover, the empty natural range has a unique representation
 corresponding to [n = p = 0]. *)

(*
val empty_natural_range : natural_range;;
(** Empty natural-dimensional range. *)

val is_empty_natural_range : natural_range -> bool;;
(** Indicator function for empty natural-dimensional range. *)
*)

val natural_range : Unatural.t -> Unatural.t -> natural_range;;
(** Construction function for natural ranges.

 Raise [Invalid_arg] when the second argument is lesser than the first. *)

(** {6 Uni-dimensional index set} *)
type index_set_1d = private
  | Is_empty
  | Is_range of natural_range
    begin
      rule Is_range { left = 0; right = 0; } -> Is_empty
    end
  | Is_extensional of index array
    begin
      rule Is_extensional [||] -> Is_empty
    end
  | Is_union of index_set * index_set
    begin
      associative
      commutative
      neutral (Is_empty)
      idempotent
    end
;;

val is_empty : index_set_1d;;
val is_range : natural_range -> index_set_1d;;
val is_extensional : index array -> index_set_1d;;
val is_union : index_set_1d * index_set_1d -> index_set_1d;;

(** {6 Multi-dimensional range} *)

type multi_range = private mono_range multi_d;;
(** The type of multi-dimensional ranges. *)

val empty_multi_range : multi_range;;
(** Empty multi-dimensional range. *)

val is_empty_multi_range : multi_range -> bool;;
(** Indicator function for empty multi-dimensional range. *)

val multi_range : mono_range list -> multi_range;;
(** Construction function for multi-dimensional ranges. *)

(** {6 Index set} *)

type index = int multi_d;;
(** The type of multi-indices. *)

type index_set = private
  | Range of multi_range
  | Extensional of index array
  | Union of index_set * index_set
;;
(** The type of set of multi-indices.
 Explicit sets of multi-indices ([Extensional] constructor) are represented by
 arrays of indices sorted in strictly increasing lexicographical order. *)

val empty_index_set : index_set;;
(** Empty set of multi-indices. *)

val is_empty_index_set : index_set -> bool;;
(** Indicator function for the empty set of multi-indices. *)

val range : multi_range -> index_set;;
val extensional : index array -> index_set;;
val union : index_set -> index_set -> index_set;;
(** Construction functions associated to each type constructors.
 When several representations would be possible, the one with the smallest
 memory weight is chosen. *)
