(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: uset.ml,v 1.1 2013/06/04 17:11:41 fclement Exp $ *)

let uset_invalid_arg fun_name msg =
  Pervasives.invalid_arg (Printf.sprintf "Uset.%s: %s" fun_name msg)
;;

type depth =
  | Universe_index
  | Set_index of depth

and 'a support_index = 'a Uuniverse.index

and 'a index =
  {
   depth : depth;
   value : 'a support_index Uuniverse.index;
  }
;;

type 'a support =
  | Universe of 'a Uuniverse.t
  | Set of 'a set

and 'a set =
  | Empty of 'a support
  | Nonempty of 'a nonempty

and 'a nonempty =
  {
   support : 'a support;
   indices : 'a support_index Uuniverse.t;
  }
;;

type 'a t = 'a set;;

let support_of_set = function
  | Empty sup -> sup
  | Nonempty { support = sup; _; } -> sup
;;

let rec universe_of_set s = universe_of_support (support_of_set s)
and universe_of_support = function
  | Universe u -> u
  | Set s -> universe_of_set s
;;

let cardinal = function
  | Empty _ -> Unatural._0_
  | Nonempty { indices = indices; _; } -> indices.Uuniverse.cardinal
;;

let cardinal_of_support = function
  | Universe u -> Uuniverse.cardinal u
  | Set s -> cardinal s
;;

let empty sup = Empty sup;;

let rec valid_depth sup depth =
  match (sup, depth) with
  | (Universe u, Universe_index) -> true
  | (Set s, Set_index depth) -> valid_depth (support_of_set s) depth
  | (Universe _, Set_index _) | (Set _, Universe_index) -> false
;;

let valid_index s { depth = depth; value = index; } =
  match s with
  | Empty _ -> false
  | Nonempty ne ->
      Uuniverse.valid_index ne.indices index &&
      valid_depth ne.support depth
;;

let check_index fun_name (s : 'a set) index =
  if not (valid_index s index) then
    uset_invalid_arg fun_name
      (Printf.sprintf "index %i out of support of %d element(s)"
         (Unatural.to_int (Uindex.to_natural index.value))
         (Unatural.to_int (cardinal s)))
;;

let singleton sup index =
  check_support_index "singleton" sup index;
  Nonempty { support = sup; indices = Uuniverse.make Unatural._1_ index; }
;;

let mem_nonempty ne index = Uuniverse.mem ne.indices index;;

let mem s index =
  match s with
  | Empty _ -> false
  | Nonempty ne -> mem_nonempty ne index
;;

let add_to_nonempty ne index =
  check_support_index "add_to_nonempty" ne.support index;
  if not (mem_nonempty ne index) then Uuniverse.add ne.indices index
;;

let choice = function
  | Empty _ -> uset_invalid_arg "choice" "empty set"
  | Nonempty ne ->
      Uuniverse.element ne.indices (Uindex.of_natural Unatural._0_)
;;

let indices = function
  | Empty _ -> [||]
  | Nonempty ne -> Uuniverse.elements ne.indices
;;

let rec element s index =
  match s with
  | Empty _ -> uset_invalid_arg "element" "empty set"
  | Nonempty { support = sup; indices = indices; } ->
      check_universe_index "element" indices index;
      let index = Uuniverse.element indices index in
      match sup with
      | Universe u -> Uuniverse.element u index
      | Set s -> element s index
;;

(* We have to understand how to deal with both index levels... *)
let elements = function
  | Empty _ -> [||]
  | Nonempty ne ->
      let rec loop index_array = function
        | Universe u -> Uuniverse.sub_elements u index_array
        | Set s ->
            match s with
            | Empty _ -> assert false (* [||]? *)
            | Nonempty { support = sup; indices = indices; } ->
                let index_array =
                  Uuniverse.sub_elements indices index_array in
                loop index_array sup in
      let index_array = Uuniverse.elements ne.indices in
      loop index_array ne.support
;;

let same_universe s1 s2 = universe_of_set s1 == universe_of_set s2;;

let check_same_universe fun_name s1 s2 =
  if not (same_universe s1 s2) then uset_invalid_arg fun_name "universe clash"
;;

let fold_left f accu = function
  | Empty _ -> accu
  | Nonempty ne -> Uuniverse.fold_left f accu ne.indices
;;

let union s1 s2 =
  check_same_universe "union" s1 s2;
  match (s1, s2) with
  | (Empty _, s) | (s, Empty _) -> s
  | (Nonempty ne1, Nonempty ne2) ->
      let (smaller, greater) =
        if cardinal s1 < cardinal s2 then (ne1, ne2) else (ne2, ne1) in
      let indices =
        fold_left
          (fun accu i -> Uuniverse.add accu i; accu)
          greater.indices (Nonempty smaller) in
      Nonempty { greater with indices = indices; }
;;

let add s i =
  let sup = support_of_set s in
  check_support_index "add" sup i;
  let single = singleton sup i in
  match s with
  | Empty _ -> single
  | Nonempty _ -> union s single
;;

let iter f = function
  | Empty _ -> ()
  | Nonempty ne -> Uuniverse.iter f ne.indices
;;

let map_indices f = function
  | Empty _ -> [||]
  | Nonempty ne -> Uuniverse.map_elements f ne.indices
;;

let of_array sup index_array =
  let len = Uarray.length index_array in
  if len = Unatural._0_ then Empty sup else
  Nonempty
    { support = sup; indices = Uuniverse.of_array index_array; }
;;

let of_list sup l = of_array sup (Array.of_list l);;

let map f s sup = of_array sup (map_indices f s);;

let find p = function
  | Empty _ -> raise Not_found
  | Nonempty ne -> Uuniverse.find p ne.indices
;;

let find_all p = function
  | Empty _ -> raise Not_found
  | Nonempty ne -> Uuniverse.find_all p ne.indices
;;

let filter = find_all;;

(*
let flatten = function
  | Empty sup -> fold_left () ...
;;
*)

let to_list = function
  | Empty _ -> []
  | Nonempty ne -> Uuniverse.to_list ne.indices
;;

(* Quotient set. *)

(*
type 'a quotient_set = 'a set list;;

let quotient_set s part =
  let max_part_index_of (m, accu) x =
    let n = part x in
    (max m n, (x, n) :: accu) in 
  let (max_part_index, rev_part_graph) =
    Ulist.fold_left max_part_index_of (Unatural._0_, []) (to_list s) in
  let nb_part = Unatural.succ max_part_index in
  let part_array = Uarray.make nb_part [] in
  Ulist.iter
    (fun (x, n) ->
      let n = Unatural.to_int n in
      part_array.(n) <- x :: part_array.(n))
    rev_part_graph;
  of_array (Array.map of_list part_array)
;;
*)

let dichotomy p s =
  let sup = Set s in
  match s with
  | Empty _ -> (Empty sup, Empty sup)
  | Nonempty { indices = indices; _; } ->
      let f (tl, fl) i =
        if p i
        then ((Uuniverse.get_index indices i) :: tl, fl)
        else (tl, (Uuniverse.get_index indices i) :: fl) in
      let (true_list, false_list) = Uuniverse.fold_left f ([], []) indices in
      (of_list sup true_list, of_list sup false_list)
;;
